@@ -36,6 +36,7 @@ from pydantic_core import CoreSchema, core_schema
 from typing_extensions import (
     Annotated,
     Literal,
+    Never,
     NotRequired,
     ParamSpec,
     TypedDict,
@@ -358,7 +359,7 @@ def test_cache_keys_are_hashable(clean_cache):
     assert len(_GENERIC_TYPES_CACHE) == cache_size + 15
 
     class Model(BaseModel):
-        x: MyGenericModel[Callable[[C], Iterable[str]]] = Field(...)
+        x: MyGenericModel[Callable[[C], Iterable[str]]]
 
     models.append(Model)
     assert len(_GENERIC_TYPES_CACHE) == cache_size + 15
@@ -475,7 +476,10 @@ def test_generics_work_with_many_parametrized_base_models(clean_cache):
         Working = B[m]
         generics.append(Working)
 
-    target_size = cache_size + count_create_models * 3 + 2
+    # A[int, C] -> 2 caches
+    # A[int, C][M] -> 3 caches  // this is generated by `mro`
+    # B[M] -> 3 caches
+    target_size = cache_size + count_create_models * 3 * 2 + 2
     assert len(_GENERIC_TYPES_CACHE) < target_size + _LIMITED_DICT_SIZE
     del models
     del generics
@@ -685,10 +689,7 @@ def test_nested():
     OuterT_SameType[int](i={'a': 8})
     OuterT_SameType[int](i=inner_int)
     OuterT_SameType[str](i=inner_str)
-    # TODO: The next line is failing, but passes in v1.
-    #   Should re-parse-from-dict if the pydantic_generic_origin is the same
-    # OuterT_SameType[str](i=inner_int_any)
-    OuterT_SameType[int](i=inner_int_any.model_dump())
+    OuterT_SameType[int](i=inner_int_any)
 
     with pytest.raises(ValidationError) as exc_info:
         OuterT_SameType[int](i=inner_str.model_dump())
@@ -706,11 +707,10 @@ def test_nested():
     # insert_assert(exc_info.value.errors(include_url=False))
     assert exc_info.value.errors(include_url=False) == [
         {
-            'type': 'model_type',
-            'loc': ('i',),
-            'msg': 'Input should be a valid dictionary or instance of InnerT[int]',
-            'input': InnerT[str](a='ate'),
-            'ctx': {'class_name': 'InnerT[int]'},
+            'type': 'int_parsing',
+            'loc': ('i', 'a'),
+            'msg': 'Input should be a valid integer, unable to parse string as an integer',
+            'input': 'ate',
         }
     ]
 
@@ -724,13 +724,7 @@ def test_nested():
         OuterT_SameType[int](i=inner_dict_any)
     # insert_assert(exc_info.value.errors(include_url=False))
     assert exc_info.value.errors(include_url=False) == [
-        {
-            'type': 'model_type',
-            'loc': ('i',),
-            'msg': 'Input should be a valid dictionary or instance of InnerT[int]',
-            'input': InnerT[Any](a={}),
-            'ctx': {'class_name': 'InnerT[int]'},
-        }
+        {'type': 'int_type', 'loc': ('i', 'a'), 'msg': 'Input should be a valid integer', 'input': {}}
     ]
 
 
@@ -2320,7 +2314,7 @@ def test_construct_generic_model_with_validation():
     class Page(BaseModel, Generic[T]):
         page: int = Field(ge=42)
         items: Sequence[T]
-        unenforced: PositiveInt = Field(..., lt=10)
+        unenforced: PositiveInt = Field(lt=10)
 
     with pytest.raises(ValidationError) as exc_info:
         Page[int](page=41, items=[], unenforced=11)
@@ -2416,7 +2410,7 @@ def test_generic_enum_bound():
 
     # insert_assert(Model[MyEnum].model_json_schema())
     assert Model[MyEnum].model_json_schema() == {
-        '$defs': {'MyEnum': {'const': 1, 'enum': [1], 'title': 'MyEnum', 'type': 'integer'}},
+        '$defs': {'MyEnum': {'enum': [1], 'title': 'MyEnum', 'type': 'integer'}},
         'properties': {'x': {'$ref': '#/$defs/MyEnum'}},
         'required': ['x'],
         'title': 'Model[test_generic_enum_bound.<locals>.MyEnum]',
@@ -2468,7 +2462,7 @@ def test_generic_intenum_bound():
 
     # insert_assert(Model[MyEnum].model_json_schema())
     assert Model[MyEnum].model_json_schema() == {
-        '$defs': {'MyEnum': {'const': 1, 'enum': [1], 'title': 'MyEnum', 'type': 'integer'}},
+        '$defs': {'MyEnum': {'enum': [1], 'title': 'MyEnum', 'type': 'integer'}},
         'properties': {'x': {'$ref': '#/$defs/MyEnum'}},
         'required': ['x'],
         'title': 'Model[test_generic_intenum_bound.<locals>.MyEnum]',
@@ -2875,3 +2869,271 @@ def test_generic_with_allow_extra():
     # This used to raise an error related to accessing the __annotations__ attribute of the Generic class
     class AllowExtraGeneric(BaseModel, Generic[T], extra='allow'):
         data: T
+
+
+def test_generic_mro_single():
+    T = TypeVar('T')
+
+    # 1.
+    class A(BaseModel, Generic[T]): ...
+
+    class B(A, Generic[T]): ...
+
+    assert B[int].__mro__ == (B[int], B, A[int], A, BaseModel, Generic, object)
+
+    # 2.
+    class A(BaseModel, Generic[T]): ...
+
+    class B(A[T]): ...
+
+    class C(B[int]): ...
+
+    assert B.__mro__ == (B, A, BaseModel, Generic, object)
+    assert C.__mro__ == (C, B[int], B, A[int], A, BaseModel, Generic, object)
+
+    # 3.
+    class A(BaseModel, Generic[T]): ...
+
+    class B(A[T], Generic[T]): ...
+
+    class C(B[T], Generic[T]): ...
+
+    assert B.__mro__ == (B, A, BaseModel, Generic, object)
+    assert B[int].__mro__ == (B[int], B, A[int], A, BaseModel, Generic, object)
+    assert C[int].__mro__ == (C[int], C, B[int], B, A[int], A, BaseModel, Generic, object)
+
+    # 4.
+    class A(BaseModel, Generic[T]): ...
+
+    class B(A[int], Generic[T]): ...
+
+    class C(B[str], Generic[T]): ...
+
+    assert B.__mro__ == (B, A[int], A, BaseModel, Generic, object)
+    assert C.__mro__ == (C, B[str], B, A[int], A, BaseModel, Generic, object)
+
+
+def test_generic_mro_multi():
+    T1 = TypeVar('T1')
+    T2 = TypeVar('T2')
+    T3 = TypeVar('T3')
+
+    # 1.
+    class A(BaseModel, Generic[T1, T2]): ...
+
+    class B(A[int, T2], BaseModel, Generic[T2]): ...
+
+    assert A[int, T2][str].__mro__ == (A[int, T2][str], A, BaseModel, Generic, object)
+    assert B[str].__mro__ == (B[str], B, A[int, T2][str], A, BaseModel, Generic, object)
+
+    # 2.
+    class A(BaseModel, Generic[T1, T2]): ...
+
+    B = A[int, T2]
+    C = B[str]
+
+    assert B.__mro__ == (A[int, T2], A, BaseModel, Generic, object)
+    assert C.__mro__ == (B[str], A, BaseModel, Generic, object)
+
+    # 3.
+
+    class A(BaseModel, Generic[T1]): ...
+
+    class B1(A[T1], Generic[T1, T2]): ...
+
+    class B2(A[T2], Generic[T1, T2]): ...
+
+    assert B1[T1, str].__mro__ == (B1[T1, str], B1, A, BaseModel, Generic, object)
+    assert B2[T1, str].__mro__ == (B2[T1, str], B2, A[str], A, BaseModel, Generic, object)
+
+    assert B1[T1, str][int].__mro__ == (B1[T1, str][int], B1, A[int], A, BaseModel, Generic, object)
+    assert B2[T1, str][int].__mro__ == (B2[T1, str][int], B2, A[str], A, BaseModel, Generic, object)
+
+    assert B1[int, str].__mro__ == (B1[int, str], B1, A[int], A, BaseModel, Generic, object)
+    assert B2[int, str].__mro__ == (B2[int, str], B2, A[str], A, BaseModel, Generic, object)
+
+    # 4.
+
+    class A(BaseModel, Generic[T1]): ...
+
+    class B(A[T2], Generic[T2]): ...
+
+    class C1(B[T3], Generic[T3, T1]): ...
+
+    class C2(B[T3], Generic[T3, T2]): ...
+
+    assert C1[int, str].__mro__ == (C1[int, str], C1, B[int], B, A[int], A, BaseModel, Generic, object)
+    assert C2[int, str].__mro__ == (C2[int, str], C2, B[int], B, A[int], A, BaseModel, Generic, object)
+
+
+def test_generic_mro_circular_typevar():
+    T1 = TypeVar('T1')
+    T2 = TypeVar('T2')
+    T3 = TypeVar('T3')
+
+    class A(BaseModel, Generic[T1]): ...
+
+    class B(A[T2], Generic[T2]): ...
+
+    class C(B[T3], Generic[T3]): ...
+
+    class D(C[T1], Generic[T1]): ...
+
+    class E(D[T2], Generic[T2]): ...
+
+    assert E[int].__mro__ == (E[int], E, D[int], D, C[int], C, B[int], B, A[int], A, BaseModel, Generic, object)
+
+
+def test_generic_mro_swap():
+    T1 = TypeVar('T1')
+    T2 = TypeVar('T2')
+
+    class A(BaseModel, Generic[T1]): ...
+
+    class B(A[T1], Generic[T1, T2]): ...
+
+    class BSwap(B[T2, T1], Generic[T1, T2]): ...
+
+    C = BSwap[T1, str]
+
+    assert C.__mro__ == (C, BSwap, B[str, T1], B, A[str], A, BaseModel, Generic, object)
+    assert C[int].__mro__ == (C[int], BSwap, B[str, int], B, A[str], A, BaseModel, Generic, object)
+
+    assert BSwap[int, str].__mro__[1:] == C[int].__mro__[1:]
+
+
+def test_generic_field():
+    # https://github.com/pydantic/pydantic/issues/10039
+
+    T = TypeVar('T')
+
+    class A(BaseModel, Generic[T]): ...
+
+    class B(A[T]): ...
+
+    class C(B[bool]): ...
+
+    class Model(BaseModel):
+        input_bool: A[bool]
+
+    Model(input_bool=C())
+
+    assert C.__mro__ == (C, B[bool], B, A[bool], A, BaseModel, Generic, object)
+    assert issubclass(C, A[bool]) and issubclass(C, A[int]) is False
+
+
+def test_generic_any_or_never() -> None:
+    T = TypeVar('T')
+
+    class GenericModel(BaseModel, Generic[T]):
+        f: Union[T, int]
+
+    any_json_schema = GenericModel[Any].model_json_schema()
+    assert any_json_schema['properties']['f'] == {'title': 'F'}  # any type
+
+    never_json_schema = GenericModel[Never].model_json_schema()
+    assert never_json_schema['properties']['f'] == {'type': 'integer', 'title': 'F'}
+
+
+def test_revalidation_against_any() -> None:
+    T = TypeVar('T')
+
+    class ResponseModel(BaseModel, Generic[T]):
+        content: T
+
+    class Product(BaseModel):
+        name: str
+        price: float
+
+    class Order(BaseModel):
+        id: int
+        product: ResponseModel[Any]
+
+    product = Product(name='Apple', price=0.5)
+    response1: ResponseModel[Any] = ResponseModel[Any](content=product)
+    response2: ResponseModel[Any] = ResponseModel(content=product)
+    response3: ResponseModel[Any] = ResponseModel[Product](content=product)
+
+    for response in response1, response2, response3:
+        order = Order(id=1, product=response)
+        assert isinstance(order.product.content, Product)
+
+
+def test_revalidation_without_explicit_parametrization() -> None:
+    """Note, this is seen in the test above as well, but is added here for thoroughness."""
+
+    T1 = TypeVar('T1', bound=BaseModel)
+
+    class InnerModel(BaseModel, Generic[T1]):
+        model: T1
+
+    T2 = TypeVar('T2', bound=InnerModel)
+
+    class OuterModel(BaseModel, Generic[T2]):
+        inner: T2
+
+    class MyModel(BaseModel):
+        foo: int
+
+    # Construct two instances, with and without generic annotation in the constructor:
+    inner1 = InnerModel[MyModel](model=MyModel(foo=42))
+    inner2 = InnerModel(model=MyModel(foo=42))
+    assert inner1 == inner2
+
+    outer1 = OuterModel[InnerModel[MyModel]](inner=inner1)
+    outer2 = OuterModel[InnerModel[MyModel]](inner=inner2)
+    # implies that validation succeeds for both
+    assert outer1 == outer2
+
+
+def test_revalidation_with_basic_inference() -> None:
+    T = TypeVar('T')
+
+    class Inner(BaseModel, Generic[T]):
+        inner: T
+
+    class Holder(BaseModel, Generic[T]):
+        inner: Inner[T]
+
+    holder1 = Holder[int](inner=Inner[int](inner=1))
+    holder2 = Holder(inner=Inner(inner=1))
+    # implies that validation succeeds for both
+    assert holder1 == holder2
+
+
+def test_generic_mro_multi_level():
+    """Pass another generic model as an arg.
+
+    See https://github.com/pydantic/pydantic/issues/11024.
+    """
+
+    T = TypeVar('T')
+
+    class GenericBaseModel(BaseModel, Generic[T]): ...
+
+    class EnumerableModel(GenericBaseModel[T]):
+        values: List[T]
+
+    T1 = TypeVar('T1')
+    T2 = TypeVar('T2')
+
+    class CombineModel(BaseModel, Generic[T1, T2]):
+        field_1: T1
+        field_2: T2
+
+    class EnumerableCombineModel(EnumerableModel[CombineModel[T1, T2]]): ...
+
+    m = EnumerableCombineModel[int, int]
+
+    mro = (
+        EnumerableCombineModel[int, int],
+        EnumerableCombineModel,
+        EnumerableModel[CombineModel[int, int]],
+        EnumerableModel,
+        GenericBaseModel[CombineModel[int, int]],
+        GenericBaseModel,
+        BaseModel,
+        Generic,
+        object,
+    )
+    assert m.__mro__ == tuple(HasRepr(repr(m)) for m in mro)
